import collections
import functools
import math
import typing

import quicktions

from .. import enumerate as _enumerate
from .. import exceptions, markups
from .. import math as _math
from .. import typings
from ..bundle import LilyPondFormatBundle
from ..duration import Duration, Multiplier, NonreducedFraction
from ..new import new
from ..ratio import Ratio
from ..sequence import Sequence
from ..storage import FormatSpecification, StorageFormatManager


@functools.total_ordering
class MetronomeMark:
    r"""
    MetronomeMark.

    ..  container:: example

        Initializes integer-valued metronome mark:

        >>> score = abjad.Score()
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark((1, 4), 90)
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \tempo 4=90
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

    ..  container:: example

        Initializes rational-valued metronome mark:

        >>> import quicktions
        >>> score = abjad.Score()
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark((1, 4), quicktions.Fraction(272, 3))
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \tempo \markup \abjad-metronome-mark-mixed-number-markup #2 #0 #1 #"90" #"2" #"3"
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

        Overrides rational-valued metronome mark with decimal string:

        >>> score = abjad.Score()
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark(
        ...     (1, 4),
        ...     quicktions.Fraction(272, 3),
        ...     decimal="90.66",
        ... )
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \tempo \markup \abjad-metronome-mark-markup #2 #0 #1 #"90.66"
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

        Overrides rational-valued metronome mark with exact decimal:

        >>> score = abjad.Score()
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark(
        ...     (1, 4),
        ...     quicktions.Fraction(901, 10),
        ...     decimal=True,
        ... )
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \tempo \markup \abjad-metronome-mark-markup #2 #0 #1 #"90.1"
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

    ..  container:: example

        Initializes from text, duration and range:

        >>> score = abjad.Score()
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark((1, 4), (120, 133), 'Quick')
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \tempo Quick 4=120-133
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_custom_markup",
        "_decimal",
        "_hide",
        "_reference_duration",
        "_textual_indication",
        "_units_per_minute",
    )

    _context = "Score"

    _format_slot = "opening"

    _mutates_offsets_in_seconds = True

    _parameter = "METRONOME_MARK"

    _persistent = True

    ### INITIALIZER ###

    def __init__(
        self,
        reference_duration: typings.DurationTyping = None,
        units_per_minute: typing.Union[int, quicktions.Fraction] = None,
        textual_indication: str = None,
        *,
        custom_markup: markups.Markup = None,
        decimal: typing.Union[bool, str] = None,
        hide: bool = None,
    ) -> None:
        assert isinstance(textual_indication, (str, type(None)))
        arguments = (reference_duration, units_per_minute, textual_indication)
        if all(_ is None for _ in arguments):
            reference_duration = (1, 4)
            units_per_minute = 60
        if reference_duration:
            reference_duration = Duration(reference_duration)
        if isinstance(units_per_minute, float):
            raise Exception(
                f"do not set units-per-minute to float ({units_per_minute});"
                " use fraction with decimal override instead."
            )
        prototype = (int, quicktions.Fraction, collections.abc.Sequence, type(None))
        assert isinstance(units_per_minute, prototype)
        if isinstance(units_per_minute, collections.abc.Sequence):
            assert len(units_per_minute) == 2
            item_prototype = (int, Duration)
            assert units_per_minute is not None
            assert all(isinstance(x, item_prototype) for x in units_per_minute)
            units_per_minute = tuple(sorted(units_per_minute))
        self._reference_duration = reference_duration
        self._textual_indication = textual_indication
        self._units_per_minute = units_per_minute
        if custom_markup is not None:
            assert isinstance(custom_markup, markups.Markup), repr(custom_markup)
        self._custom_markup = custom_markup
        if decimal is not None:
            assert isinstance(decimal, (bool, str)), repr(decimal)
        self._decimal = decimal
        if hide is not None:
            hide = bool(hide)
        self._hide = hide

    ### SPECIAL METHODS ###

    def __add__(self, argument) -> typing.Optional["MetronomeMark"]:
        """
        Adds metronome mark to ``argument``.

        ..  container:: example

            Adds one metronome mark to another:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 60)
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

            >>> mark_2 + mark_1
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

        ..  container:: example exception

            Raises imprecise metronome mark error with textual indication:

            >>> mark_1 = abjad.MetronomeMark(textual_indication='Langsam')
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example exception

            Raises imprecise metronome mark error with range:

            >>> mark_1 = abjad.MetronomeMark((1, 8), (90, 92))
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example exception

            Raises type error when ``argument`` is not a metronome mark:

            >>> abjad.MetronomeMark((1, 4), 60) + 90
            Traceback (most recent call last):
                ...
            TypeError: 90

        """
        if not isinstance(argument, type(self)):
            raise TypeError(argument)
        if self.is_imprecise or argument.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.quarters_per_minute, quicktions.Fraction)
        assert isinstance(argument.quarters_per_minute, quicktions.Fraction)
        assert isinstance(self.reference_duration, Duration)
        assert isinstance(argument.reference_duration, Duration)
        new_quarters_per_minute = (
            self.quarters_per_minute + argument.quarters_per_minute
        )
        minimum_denominator = min(
            (
                self.reference_duration.denominator,
                argument.reference_duration.denominator,
            )
        )
        nonreduced_fraction = NonreducedFraction(new_quarters_per_minute / 4)
        nonreduced_fraction = nonreduced_fraction.with_denominator(minimum_denominator)
        (
            new_units_per_minute,
            new_reference_duration_denominator,
        ) = nonreduced_fraction.pair
        new_reference_duration = Duration(1, new_reference_duration_denominator)
        metronome_mark = type(self)(new_reference_duration, new_units_per_minute)
        return metronome_mark

    def __div__(self, argument) -> "MetronomeMark":
        """
        Divides metronome mark by ``argument``.

        ..  container:: example

            Divides metronome mark by number:

            >>> abjad.MetronomeMark((1, 4), 60) / 2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=30)

        ..  container:: example

            Divides metronome mark by other metronome mark:

            >>> abjad.MetronomeMark((1, 4), 60) / abjad.MetronomeMark((1, 4), 40)
            Multiplier(3, 2)

        """
        if self.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        if getattr(argument, "is_imprecise", False):
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.quarters_per_minute, quicktions.Fraction)
        if isinstance(argument, type(self)):
            assert isinstance(argument.quarters_per_minute, quicktions.Fraction)
            result = self.quarters_per_minute / argument.quarters_per_minute
            return Multiplier(result)
        elif isinstance(argument, (int, quicktions.Fraction)):
            assert isinstance(self.units_per_minute, (int, quicktions.Fraction))
            units_per_minute = self.units_per_minute / argument
            if _math.is_integer_equivalent_number(units_per_minute):
                units_per_minute = int(units_per_minute)
            else:
                units_per_minute = quicktions.Fraction(units_per_minute)
            result = new(self, units_per_minute=units_per_minute)
            return result
        else:
            raise TypeError(f"must be number or metronome mark: {argument!r}.")

    def __eq__(self, argument) -> bool:
        """
        Is true when metronome mark equals ``argument``.

        ..  container:: example

            >>> mark_1 = abjad.MetronomeMark((3, 32), 52)
            >>> mark_2 = abjad.MetronomeMark((3, 32), 52)

            >>> mark_1 == mark_2
            True

            >>> mark_2 == mark_1
            True

        ..  container:: example

            >>> mark_1 = abjad.MetronomeMark((3, 32), 52)
            >>> mark_2 = abjad.MetronomeMark((6, 32), 104)

            >>> mark_1 == mark_2
            False

            >>> mark_2 == mark_1
            False

        ..  container:: example

            >>> mark_1 = abjad.MetronomeMark((3, 32), 52, 'Langsam')
            >>> mark_2 = abjad.MetronomeMark((3, 32), 52, 'Langsam')
            >>> mark_3 = abjad.MetronomeMark((3, 32), 52, 'Slow')

            >>> mark_1 == mark_2
            True

            >>> mark_1 == mark_3
            False

        """
        if not isinstance(argument, type(self)):
            return False
        # 'hide' is excluded from equality testing:
        if (
            self.reference_duration == argument.reference_duration
            and self.units_per_minute == argument.units_per_minute
            and self.textual_indication == argument.textual_indication
            and self.custom_markup == argument.custom_markup
        ):
            return True
        return False

    def __hash__(self) -> int:
        """
        Hashes metronome mark.

        Redefined in tandem with __eq__.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __lt__(self, argument) -> bool:
        """
        Is true when ``argument`` is a metronome mark with quarters per
        minute greater than that of this metronome mark.
        """
        assert isinstance(argument, type(self)), repr(argument)
        self_quarters_per_minute = self.quarters_per_minute or 0
        argument_quarters_per_minute = argument.quarters_per_minute or 0
        assert isinstance(self_quarters_per_minute, (int, float, quicktions.Fraction))
        assert isinstance(
            argument_quarters_per_minute, (int, float, quicktions.Fraction)
        )
        return self_quarters_per_minute < argument_quarters_per_minute

    def __mul__(
        self, multiplier: typing.Union[int, quicktions.Fraction]
    ) -> typing.Optional["MetronomeMark"]:
        """
        Multiplies metronome mark by ``multiplier``.

        ..  container:: example

            Doubles metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> 2 * mark
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=168)

        ..  container:: example

            Triples metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> 3 * mark
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=252)

        """
        if not isinstance(multiplier, (int, quicktions.Fraction)):
            return None
        if self.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.units_per_minute, (int, quicktions.Fraction))
        new_units_per_minute = multiplier * self.units_per_minute
        new_reference_duration = Duration(self.reference_duration)
        metronome_mark = type(self)(
            reference_duration=new_reference_duration,
            units_per_minute=new_units_per_minute,
        )
        return metronome_mark

    def __radd__(self, argument) -> typing.Optional["MetronomeMark"]:
        """
        Adds ``argument`` to metronome mark.

        ..  container:: example

            Adds one metronome mark to another:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 60)
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

            >>> mark_2 + mark_1
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

        ..  container:: example exception

            Raises imprecise metronome mark error with textual indication:

            >>> mark_1 = abjad.MetronomeMark(textual_indication='Langsam')
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example exception

            Raises imprecise metronome mark error with range:

            >>> mark_1 = abjad.MetronomeMark((1, 8), (90, 92))
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example exception

            Raises type error when ``argument`` is not a metronome mark:

            >>> 90 + abjad.MetronomeMark((1, 4), 60)
            Traceback (most recent call last):
                ...
            TypeError: 90

        """
        if not isinstance(argument, type(self)):
            raise TypeError(argument)
        return argument.__add__(self)

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __rmul__(
        self, multiplier: typing.Union[int, quicktions.Fraction]
    ) -> typing.Optional["MetronomeMark"]:
        """
        Multiplies ``multiplier`` by metronome mark.

        ..  container::: example

            Doubles metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> mark * 2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=168)

        ..  container::: example

            Triples metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> mark * 3
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=252)

        """
        if not isinstance(multiplier, (int, quicktions.Fraction)):
            return None
        if self.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.units_per_minute, (int, quicktions.Fraction))
        new_units_per_minute = multiplier * self.units_per_minute
        new_reference_duration = Duration(self.reference_duration)
        metronome_mark = type(self)(
            reference_duration=new_reference_duration,
            units_per_minute=new_units_per_minute,
        )
        return metronome_mark

    def __str__(self) -> str:
        """
        Gets string representation of metronome mark.

        ..  container:: example

            Integer-valued metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> str(mark)
            '4=90'

        ..  container:: example

            Rational-valued metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), (272, 3))
            >>> str(mark)
            '4=3-272'

        """
        if self.textual_indication is not None:
            string = self.textual_indication
        elif isinstance(self.units_per_minute, (int, float)):
            string = f"{self._dotted}={self.units_per_minute}"
        elif isinstance(
            self.units_per_minute, quicktions.Fraction
        ) and not _math.is_integer_equivalent_number(self.units_per_minute):
            integer_part = int(float(self.units_per_minute))
            remainder = self.units_per_minute - integer_part
            remainder = quicktions.Fraction(remainder)
            string = f"{self._dotted}={integer_part}+{remainder}"
        elif isinstance(
            self.units_per_minute, quicktions.Fraction
        ) and _math.is_integer_equivalent_number(self.units_per_minute):
            integer = int(float(self.units_per_minute))
            string = f"{self._dotted}={integer}"
        elif isinstance(self.units_per_minute, tuple):
            first = self.units_per_minute[0]
            second = self.units_per_minute[1]
            string = f"{self._dotted}={first}-{second}"
        else:
            raise TypeError(f"unknown: {self.units_per_minute!r}.")
        return string

    def __sub__(self, argument) -> "MetronomeMark":
        """
        Subtracts ``argument`` from metronome mark.

        ..  container:: example

            Same reference reference durations:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_2 = abjad.MetronomeMark((1, 4), 60)
            >>> mark_1 - mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=30)

        ..  container:: example

            Different reference durations:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_2 = abjad.MetronomeMark((1, 2), 90)
            >>> mark_1 - mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=45)

        ..  container:: example exception

            Raises imprecise metronome mark error with textual indication:

            >>> mark_1 = abjad.MetronomeMark(textual_indication='Langsam')
            >>> mark_2 = abjad.MetronomeMark((1, 2), 90)
            >>> mark_1 - mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        """
        if not isinstance(argument, type(self)):
            raise Exception("must be metronome mark: {argument!r}.")
        if self.is_imprecise or argument.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.quarters_per_minute, (int, float, quicktions.Fraction))
        assert isinstance(
            argument.quarters_per_minute, (int, float, quicktions.Fraction)
        )
        assert isinstance(self.reference_duration, Duration)
        assert isinstance(argument.reference_duration, Duration)
        new_quarters_per_minute = (
            self.quarters_per_minute - argument.quarters_per_minute
        )
        minimum_denominator = min(
            (
                self.reference_duration.denominator,
                argument.reference_duration.denominator,
            )
        )
        nonreduced_fraction = NonreducedFraction(new_quarters_per_minute / 4)
        nonreduced_fraction = nonreduced_fraction.with_denominator(minimum_denominator)
        (
            new_units_per_minute,
            new_reference_duration_denominator,
        ) = nonreduced_fraction.pair
        new_reference_duration = Duration(1, new_reference_duration_denominator)
        metronome_mark = type(self)(
            reference_duration=new_reference_duration,
            units_per_minute=new_units_per_minute,
        )
        return metronome_mark

    def __truediv__(self, argument) -> "MetronomeMark":
        """
        Divides metronome mark by ``argument``.

        Operator required by Python 3.

        ..  container:: example

            Divides metronome mark by number:

            >>> abjad.MetronomeMark((1, 4), 60).__truediv__(2)
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=30)

        ..  container:: example

            Divides metronome mark by other metronome mark:

            >>> abjad.MetronomeMark((1, 4), 60).__truediv__(
            ...     abjad.MetronomeMark((1, 4), 40)
            ...     )
            Multiplier(3, 2)

        """
        return self.__div__(argument)

    ### PRIVATE PROPERTIES ###

    @property
    def _dotted(self):
        return self.reference_duration.lilypond_duration_string

    @property
    def _equation(self):
        # TODO: remove this assert after running all tests:
        assert not isinstance(self.units_per_minute, float)
        if self.reference_duration is None:
            return
        if isinstance(self.units_per_minute, tuple):
            first, second = self.units_per_minute
            string = f"{self._dotted}={first}-{second}"
            return string
        elif isinstance(self.units_per_minute, quicktions.Fraction):
            markup = MetronomeMark.make_tempo_equation_markup(
                self.reference_duration,
                self.units_per_minute,
                decimal=self.decimal,
            )
            string = str(markup)
            return string
        string = f"{self._dotted}={self.units_per_minute}"
        return string

    def _get_format_specification(self):
        return FormatSpecification(client=self)

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        text, equation = None, None
        if self.textual_indication is not None:
            text = self.textual_indication
            assert isinstance(text, str)
            if " " in text:
                text = f'"{text}"'
        if self.reference_duration is not None and self.units_per_minute is not None:
            equation = self._equation
        if self.custom_markup is not None:
            return rf"\tempo {self.custom_markup}"
        elif text and equation:
            return rf"\tempo {text} {equation}"
        elif equation:
            return rf"\tempo {equation}"
        elif text:
            return rf"\tempo {text}"
        else:
            return r"\tempo \default"

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if not self.hide:
            bundle.before.commands.append(self._get_lilypond_format())
        return bundle

    def _get_markup(self):
        if self.custom_markup is not None:
            return self.custom_markup
        duration_log = int(math.log(self.reference_duration.denominator, 2))
        stem_height = 1
        markup = markups.abjad_metronome_mark(
            duration_log,
            self.reference_duration.dot_count,
            stem_height,
            self.units_per_minute,
        )
        return markup

    # TODO: refactor to return dict
    def _get_markup_arguments(self):
        assert self.custom_markup is None
        duration_log = int(math.log(self.reference_duration.denominator, 2))
        dot_count = self.reference_duration.dot_count
        stem_height = 1
        if not self.decimal:
            return (
                duration_log,
                dot_count,
                stem_height,
                self.units_per_minute,
            )
        if isinstance(self.decimal, str):
            return (duration_log, dot_count, stem_height, self.decimal)
        assert self.decimal is True, repr(self.decimal)
        # TODO: add abjad.NonreducedFraction.mixed_number property
        fraction = NonreducedFraction(self.units_per_minute)
        n, d = fraction.pair
        base = n // d
        n = n % d
        return (duration_log, dot_count, stem_height, base, n, d)

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Gets (historically conventional) context.

        ..  container:: example

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.context
            'Score'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def custom_markup(self) -> typing.Optional[markups.Markup]:
        r"""
        Gets custom markup of metronome mark.

        ..  container:: example

            With custom markup:

            >>> import quicktions
            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     abjad.Duration(1, 4),
            ...     67.5,
            ...  )
            >>> mark = abjad.MetronomeMark(
            ...     reference_duration=(1, 4),
            ...     units_per_minute=quicktions.Fraction(135, 2),
            ...     custom_markup=markup,
            ...  )
            >>> staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> score = abjad.Score([staff])
            >>> abjad.attach(mark, staff[0])
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \new Score
                <<
                    \new Staff
                    {
                        \tempo \markup \abjad-metronome-mark-markup #2 #0 #1 #"67.5"
                        c'4
                        d'4
                        e'4
                        f'4
                    }
                >>

        """
        return self._custom_markup

    @property
    def decimal(self) -> typing.Union[bool, str, None]:
        """
        Gets decimal override.

        ..  container:: example

            >>> import quicktions
            >>> mark = abjad.MetronomeMark(
            ...     (1, 4),
            ...     quicktions.Fraction(272, 3),
            ... )
            >>> mark.decimal is None
            True

            >>> mark = abjad.MetronomeMark(
            ...     (1, 4),
            ...     quicktions.Fraction(272, 3),
            ...     decimal="90.66",
            ... )
            >>> mark.decimal
            '90.66'

            >>> mark = abjad.MetronomeMark(
            ...     (1, 4),
            ...     quicktions.Fraction(901, 10),
            ...     decimal=True,
            ... )
            >>> mark.decimal
            True

        """
        return self._decimal

    @property
    def hide(self) -> typing.Optional[bool]:
        r"""
        Is true when metronome mark should not appear in output (but
        should still determine effective metronome mark).

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> metronome_mark_1 = abjad.MetronomeMark((1, 4), 72)
            >>> abjad.attach(metronome_mark_1, staff[0])
            >>> metronome_mark_2 = abjad.MetronomeMark(
            ...     textual_indication='Allegro',
            ...     hide=True,
            ... )
            >>> abjad.attach(metronome_mark_2, staff[2])
            >>> score = abjad.Score([staff])
            >>> abjad.show(score) # doctest: +SKIP

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \tempo 4=72
                    c'4
                    d'4
                    e'4
                    f'4
                }
            >>

            >>> for leaf in abjad.iterate(staff).leaves():
            ...     prototype = abjad.MetronomeMark
            ...     leaf, abjad.get.effective(leaf, prototype)
            ...
            (Note("c'4"), MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=72))
            (Note("d'4"), MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=72))
            (Note("e'4"), MetronomeMark(textual_indication='Allegro', hide=True))
            (Note("f'4"), MetronomeMark(textual_indication='Allegro', hide=True))

        """
        return self._hide

    @property
    def is_imprecise(self) -> bool:
        """
        Is true if metronome mark is entirely textual or if metronome
        mark's units_per_minute is a range.

        ..  container:: example

            Imprecise metronome marks:

            >>> abjad.MetronomeMark((1, 4), 60).is_imprecise
            False
            >>> abjad.MetronomeMark(4, 60, 'Langsam').is_imprecise
            False
            >>> abjad.MetronomeMark(textual_indication='Langsam').is_imprecise
            True
            >>> abjad.MetronomeMark(4, (35, 50), 'Langsam').is_imprecise
            True
            >>> abjad.MetronomeMark((1, 4), (35, 50)).is_imprecise
            True

        ..  container:: example

            Precise metronome marks:

            >>> abjad.MetronomeMark((1, 4), 60).is_imprecise
            False

        """
        if self.reference_duration is not None:
            if self.units_per_minute is not None:
                if not isinstance(self.units_per_minute, tuple):
                    return False
        return True

    @property
    def parameter(self) -> str:
        """
        Is ``'METRONOME_MARK'``.

        ..  container:: example

            >>> abjad.MetronomeMark((1, 8), 52).parameter
            'METRONOME_MARK'

        """
        return self._parameter

    @property
    def persistent(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.MetronomeMark((1, 8), 52).persistent
            True

        """
        return self._persistent

    @property
    def quarters_per_minute(self) -> typing.Union[tuple, None, quicktions.Fraction]:
        """
        Gets metronome mark quarters per minute.

        ..  container:: example

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.quarters_per_minute
            Fraction(104, 1)

        Gives tuple when metronome mark ``units_per_minute`` is a range.

        Gives none when metronome mark is imprecise.

        Gives fraction otherwise.
        """
        if self.is_imprecise:
            return None
        if isinstance(self.units_per_minute, tuple):
            low = Duration(1, 4) / self.reference_duration * self.units_per_minute[0]
            high = Duration(1, 4) / self.reference_duration * self.units_per_minute[1]
            return (low, high)
        result = Duration(1, 4) / self.reference_duration * self.units_per_minute
        return quicktions.Fraction(result)

    @property
    def reference_duration(self) -> typing.Optional[Duration]:
        """
        Gets reference duration of metronome mark.

        ..  container:: example

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.reference_duration
            Duration(1, 8)

        """
        return self._reference_duration

    @property
    def textual_indication(self) -> typing.Optional[str]:
        """
        Gets optional textual indication of metronome mark.

        ..  container:: example

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.textual_indication is None
            True

        """
        return self._textual_indication

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on metronome mark.

        The LilyPond ``\tempo`` command refuses tweaks.

        Override the LilyPond ``MetronomeMark`` grob instead.
        """
        pass

    @property
    def units_per_minute(self) -> typing.Union[int, quicktions.Fraction, None]:
        """
        Gets units per minute of metronome mark.

        ..  container:: example

            Integer-valued metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> mark.units_per_minute
            90

        ..  container:: example

            Rational-valued metronome mark:

            >>> import quicktions
            >>> mark = abjad.MetronomeMark((1, 4), quicktions.Fraction(272, 3))
            >>> mark.units_per_minute
            Fraction(272, 3)

        """
        return self._units_per_minute

    ### PUBLIC METHODS ###

    def duration_to_milliseconds(self, duration) -> Duration:
        """
        Gets millisecond value of ``duration`` under a given metronome mark.

        ..  container:: example

            Dotted sixteenth lasts 1500 msec at quarter equals 60:

            >>> mark = abjad.MetronomeMark((1, 4), 60)
            >>> mark.duration_to_milliseconds((3, 8))
            Duration(1500, 1)

        """
        assert isinstance(self.reference_duration, Duration)
        denominator = self.reference_duration.denominator
        numerator = self.reference_duration.numerator
        whole_note_duration = 1000
        whole_note_duration *= Multiplier(denominator, numerator)
        whole_note_duration *= Multiplier(60, self.units_per_minute)
        duration = Duration(duration)
        return Duration(duration * whole_note_duration)

    def list_related_tempos(
        self,
        maximum_numerator=None,
        maximum_denominator=None,
        integer_tempos_only=False,
    ) -> typing.List[typing.Tuple["MetronomeMark", "Ratio"]]:
        r"""
        Lists related tempos.

        ..  container:: example

            Rewrites tempo ``4=58`` by ratios ``n:d`` such that ``1 <= n <= 8``
            and ``1 <= d <= 8``.

            >>> mark = abjad.MetronomeMark((1, 4), 58)
            >>> pairs = mark.list_related_tempos(
            ...     maximum_numerator=8,
            ...     maximum_denominator=8,
            ...  )

            >>> for tempo, ratio in pairs:
            ...     string = f'{tempo!s}    {ratio!s}'
            ...     print(string)
            4=29    1:2
            4=33+1/7    4:7
            4=34+4/5    3:5
            4=36+1/4    5:8
            4=38+2/3    2:3
            4=41+3/7    5:7
            4=43+1/2    3:4
            4=46+2/5    4:5
            4=48+1/3    5:6
            4=49+5/7    6:7
            4=50+3/4    7:8
            4=58    1:1
            4=66+2/7    8:7
            4=67+2/3    7:6
            4=69+3/5    6:5
            4=72+1/2    5:4
            4=77+1/3    4:3
            4=81+1/5    7:5
            4=87    3:2
            4=92+4/5    8:5
            4=96+2/3    5:3
            4=101+1/2    7:4
            4=116    2:1

        ..  container:: example

            Integer-valued tempos only:

            >>> mark = abjad.MetronomeMark((1, 4), 58)
            >>> pairs = mark.list_related_tempos(
            ...     maximum_numerator=16,
            ...     maximum_denominator=16,
            ...     integer_tempos_only=True,
            ...  )

            >>> for tempo, ratio in pairs:
            ...     string = f'{tempo!s}    {ratio!s}'
            ...     print(string)
            4=29    1:2
            4=58    1:1
            4=87    3:2
            4=116    2:1

        Constrains ratios such that ``1:2 <= n:d <= 2:1``.
        """
        allowable_numerators = range(1, maximum_numerator + 1)
        allowable_denominators = range(1, maximum_denominator + 1)
        numbers = [allowable_numerators, allowable_denominators]
        pairs = _enumerate.yield_outer_product(numbers)
        multipliers = [Multiplier(_) for _ in pairs]
        multipliers = [
            _
            for _ in multipliers
            if quicktions.Fraction(1, 2) <= _ <= quicktions.Fraction(2)
        ]
        multipliers.sort()
        multipliers_ = Sequence(multipliers).remove_repeats()
        pairs = []
        for multiplier in multipliers_:
            new_units_per_minute = multiplier * self.units_per_minute
            if integer_tempos_only and not _math.is_integer_equivalent_number(
                new_units_per_minute
            ):
                continue
            metronome_mark = type(self)(
                reference_duration=self.reference_duration,
                units_per_minute=new_units_per_minute,
            )
            ratio = Ratio(multiplier.pair)
            pair = (metronome_mark, ratio)
            pairs.append(pair)
        return pairs

    @staticmethod
    def make_tempo_equation_markup(
        reference_duration, units_per_minute, *, decimal=None
    ) -> markups.Markup:
        r"""
        Makes tempo equation markup.

        ..  container:: example

            Integer-valued metronome mark:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     (1, 4),
            ...     90,
            ...  )
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(abjad.lilypond(markup))
                \markup \abjad-metronome-mark-markup #2 #0 #1 #"90"

        ..  container:: example

            Float-valued metronome mark:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     (1, 4),
            ...     90.1,
            ... )
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(abjad.lilypond(markup))
                \markup \abjad-metronome-mark-markup #2 #0 #1 #"90.1"

        ..  container:: example

            Rational-valued metronome mark:

            >>> import quicktions
            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     abjad.Duration(1, 4),
            ...     quicktions.Fraction(272, 3),
            ... )
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(abjad.lilypond(markup))
                \markup \abjad-metronome-mark-mixed-number-markup #2 #0 #1 #"90" #"2" #"3"

        """
        reference_duration_ = Duration(reference_duration)
        log = reference_duration_.exponent
        dots = reference_duration_.dot_count
        stem = 1
        if isinstance(
            units_per_minute, quicktions.Fraction
        ) and not _math.is_integer_equivalent_number(units_per_minute):
            if decimal:
                decimal_: typing.Union[float, str]
                if decimal is True:
                    decimal_ = float(units_per_minute)
                else:
                    assert isinstance(decimal, str), repr(decimal)
                    decimal_ = decimal
                markup = markups.Markup(
                    r"\markup \abjad-metronome-mark-markup"
                    f' #{log} #{dots} #{stem} #"{decimal_}"',
                    literal=True,
                )
            else:
                nonreduced = NonreducedFraction(units_per_minute)
                base = int(nonreduced)
                remainder = nonreduced - base
                n, d = remainder.pair
                markup = markups.Markup(
                    r"\markup \abjad-metronome-mark-mixed-number-markup"
                    f" #{log} #{dots} #{stem}"
                    f' #"{base}" #"{n}" #"{d}"',
                    literal=True,
                )
        else:
            markup = markups.Markup(
                r"\markup \abjad-metronome-mark-markup"
                f' #{log} #{dots} #{stem} #"{units_per_minute}"',
                literal=True,
            )
        return markup

    def rewrite_duration(self, duration, metronome_mark) -> Duration:
        r"""
        Rewrites ``duration`` under ``metronome_mark``.

        ..  container:: example

            Consider the two metronome marks below.

            >>> tempo = abjad.MetronomeMark((1, 4), 60)
            >>> metronome_mark = abjad.MetronomeMark((1, 4), 90)

            ``tempo`` specifies quarter equal to ``60``.

            ``metronome_mark`` indication specifies quarter equal to ``90``.

            ``metronome_mark`` is ``3/2`` times as fast as ``tempo``:

            >>> metronome_mark / tempo
            Multiplier(3, 2)

            Note that a triplet eighth note under ``tempo`` equals a regular
            eighth note under ``metronome_mark``:

            >>> tempo.rewrite_duration((1, 12), metronome_mark)
            Duration(1, 8)

            And note that a regular eighth note under ``tempo`` equals a dotted
            sixteenth under ``metronome_mark``:

            >>> tempo.rewrite_duration((1, 8), metronome_mark)
            Duration(3, 16)

        Given ``duration`` governed by this tempo returns new duration governed
        by ``metronome_mark``.

        Ensures that ``duration`` and new duration consume the same amount of
        time in seconds.
        """
        duration = Duration(duration)
        tempo_ratio = metronome_mark / self
        new_duration = tempo_ratio * duration
        return new_duration



import functools
import numbers
import typing

from ..storage import FormatSpecification, StorageFormatManager
from . import _lib
from .Octave import Octave
from .pitchclasses import NamedPitchClass
from .pitches import NamedPitch, NumberedPitch, Pitch


@functools.total_ordering
class PitchRange:
    r"""
    Pitch range.

    ..  container:: example

        Pitches from C3 to C7, inclusive:

        >>> pitch_range = abjad.PitchRange("[C3, C7]")
        >>> lilypond_file = abjad.illustrate(pitch_range)
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.storage(pitch_range)
            >>> print(string)
            abjad.PitchRange('[C3, C7]')

    ..  container:: example

        >>> pitch_range = abjad.PitchRange("[C3, C7]")
        >>> lilypond_file = abjad.illustrate(pitch_range)
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(lilypond_file[abjad.Score])
            >>> print(string)
            \new Score
            \with
            {
                \override BarLine.stencil = ##f
                \override Glissando.thickness = 2
                \override SpanBar.stencil = ##f
                \override TimeSignature.stencil = ##f
            }
            <<
                \new PianoStaff
                <<
                    \context Staff = "Treble_Staff"
                    {
                        \clef "treble"
                        s1 * 1/4
                        s1 * 1/4
                    }
                    \context Staff = "Bass_Staff"
                    {
                        \clef "bass"
                        c1 * 1/4
                        \glissando
                        \change Staff = Treble_Staff
                        c''''1 * 1/4
                    }
                >>
            >>

    Initalizes from pitch numbers, pitch names, pitch instances,
    one-line reprs or other pitch range objects.

    Pitch ranges do not sort relative to other pitch ranges.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_start", "_stop")

    ### INITIALIZER ###

    def __init__(self, range_string="[A0, C8]"):
        if isinstance(range_string, type(self)):
            range_string = range_string.range_string
        start, stop = self._parse_range_string(range_string)
        self._start = start
        self._stop = stop

    ### SPECIAL METHODS ###

    def __contains__(self, argument) -> bool:
        """
        Is true when pitch range contains ``argument``.

        ..  container:: example

            Closed / closed range:

            >>> range_ = abjad.PitchRange("[A0, C8]")

            >>> -99 in range_
            False

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            False

        ..  container:: example

            Closed / open range:

            >>> range_ = abjad.PitchRange("[A0, C8)")

            >>> -99 in range_
            False

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            False

            >>> 99 in range_
            False

        ..  container:: example

            Closed / infinite range:

            >>> range_ = abjad.PitchRange("[-39, +inf]")

            >>> -99 in range_
            False

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            True

        ..  container:: example

            Open / closed range:

            >>> range_ = abjad.PitchRange("(A0, C8]")

            >>> -99 in range_
            False

            >>> -39 in range_
            False

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            False

        ..  container:: example

            Open / open range:

            >>> range_ = abjad.PitchRange("(A0, C8)")

            >>> -99 in range_
            False

            >>> -39 in range_
            False

            >>> 0 in range_
            True

            >>> 48 in range_
            False

            >>> 99 in range_
            False

        ..  container:: example

            Infinite / closed range:

            >>> range_ = abjad.PitchRange("[-inf, C8]")

            >>> -99 in range_
            True

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            False

        ..  container:: example

            Infinite / open range:

            >>> range_ = abjad.PitchRange("[-inf, C8)")

            >>> -99 in range_
            True

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            False

            >>> 99 in range_
            False

        ..  container:: example

            Infinite / infinite range:

            >>> range_ = abjad.PitchRange("[-inf, +inf]")

            >>> -99 in range_
            True

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            True

        """
        if isinstance(argument, (int, float)):
            pitch = NamedPitch(argument)
            return self._contains_pitch(pitch)
        if isinstance(argument, Pitch):
            return self._contains_pitch(argument)
        raise Exception(f"must be pitch or number {argument!r}.")

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a pitch range with start and stop equal
        to those of this pitch range.

        ..  container:: example

            >>> range_1 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_2 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_3 = abjad.PitchRange.from_pitches(-39, 48)

            >>> range_1 == range_1
            True
            >>> range_1 == range_2
            True
            >>> range_1 == range_3
            False

            >>> range_2 == range_1
            True
            >>> range_2 == range_2
            True
            >>> range_2 == range_3
            False

            >>> range_3 == range_1
            False
            >>> range_3 == range_2
            False
            >>> range_3 == range_3
            True

        Returns true or false.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self):
        """
        Hashes pitch range.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __lt__(self, argument):
        """
        Is true when start pitch of this pitch-range is less than start
        pitch of ``argument`` pitch range.

        ..  container:: example

            >>> range_1 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_2 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_3 = abjad.PitchRange.from_pitches(-39, 48)

            >>> range_1 < range_1
            False
            >>> range_1 < range_2
            False
            >>> range_1 < range_3
            True

            >>> range_2 < range_1
            False
            >>> range_2 < range_2
            False
            >>> range_2 < range_3
            True

            >>> range_3 < range_1
            False
            >>> range_3 < range_2
            False
            >>> range_3 < range_3
            False

        Returns true or false.
        """
        try:
            argument = type(self)(argument)
        except (TypeError, ValueError):
            return False
        if self.start_pitch == argument.start_pitch:
            return self.stop_pitch < argument.stop_pitch
        return self.start_pitch < argument.start_pitch

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE PROPERTIES ###

    @property
    def _close_bracket_string(self):
        if self.stop_pitch_is_included_in_range:
            return "]"
        else:
            return ")"

    @property
    def _open_bracket_string(self):
        if self.start_pitch_is_included_in_range:
            return "["
        else:
            return "("

    ### PRIVATE METHODS ###

    def _contains_pitch(self, pitch):
        if isinstance(pitch, numbers.Number):
            pitch = NamedPitch(pitch)
        elif isinstance(pitch, str):
            pitch = NamedPitch(pitch)
        if self.start_pitch is None and self.stop_pitch is None:
            return True
        elif self.start_pitch is None:
            if self.stop_pitch_is_included_in_range:
                return pitch <= self.stop_pitch
            else:
                return pitch < self.stop_pitch
        elif self.stop_pitch is None:
            if self.start_pitch_is_included_in_range:
                return self.start_pitch <= pitch
            else:
                return self.start_pitch < pitch
        else:
            if self.start_pitch_is_included_in_range:
                if self.stop_pitch_is_included_in_range:
                    return self.start_pitch <= pitch <= self.stop_pitch
                else:
                    return self.start_pitch <= pitch < self.stop_pitch
            else:
                if self.stop_pitch_is_included_in_range:
                    return self.start_pitch < pitch <= self.stop_pitch
                else:
                    return self.start_pitch < pitch < self.stop_pitch

    def _get_format_specification(self):
        return FormatSpecification(
            self,
            coerce_for_equality=True,
            repr_is_indented=False,
            storage_format_args_values=[self.range_string],
            storage_format_is_indented=False,
        )

    def _get_named_range_string(self):
        result = []
        result.append(self._open_bracket_string)
        if self.start_pitch:
            result.append(self.start_pitch.get_name(locale="us"))
        else:
            result.append("-inf")
        result.append(", ")
        if self.stop_pitch:
            result.append(self.stop_pitch.get_name(locale="us"))
        else:
            result.append("+inf")
        result.append(self._close_bracket_string)
        result = "".join(result)
        return result

    def _get_numbered_range_string(self):
        result = []
        result.append(self._open_bracket_string)
        result.append(str(self.start_pitch.number))
        result.append(", ")
        result.append(str(self.stop_pitch.number))
        result.append(self._close_bracket_string)
        result = "".join(result)
        return result

    def _parse_range_string(self, range_string):
        assert isinstance(range_string, str), repr(range_string)
        range_string = range_string.replace("-inf", "-1000")
        range_string = range_string.replace("+inf", "1000")
        match = _lib._range_string_regex.match(range_string)
        if match is None:
            raise ValueError(f"can not instantiate pitch range: {range_string!r}")
        group_dict = match.groupdict()
        start_punctuation = group_dict["open_bracket"]
        start_pitch_string = group_dict["start_pitch"]
        stop_pitch_string = group_dict["stop_pitch"]
        stop_punctuation = group_dict["close_bracket"]
        start_inclusivity_string = _lib._start_punctuation_to_inclusivity_string[
            start_punctuation
        ]
        stop_inclusivity_string = _lib._stop_punctuation_to_inclusivity_string[
            stop_punctuation
        ]
        if start_pitch_string == "-1000":
            start_pitch = None
        else:
            try:
                start_pitch = NamedPitch(start_pitch_string)
            except (TypeError, ValueError):
                start_pitch = NumberedPitch(int(start_pitch_string))
        if stop_pitch_string == "1000":
            stop_pitch = None
        else:
            try:
                stop_pitch = NamedPitch(stop_pitch_string)
            except (TypeError, ValueError):
                stop_pitch = NumberedPitch(int(stop_pitch_string))
        start_pair = (start_pitch, start_inclusivity_string)
        stop_pair = (stop_pitch, stop_inclusivity_string)
        return start_pair, stop_pair

    ### PUBLIC PROPERTIES ###

    @property
    def range_string(self) -> str:
        """
        Gets range string of pitch range.

        ..  container:: example

            >>> pitch_range = abjad.PitchRange("[C3, C7]")
            >>> pitch_range.range_string
            '[C3, C7]'

        """
        return self._get_named_range_string()

    @property
    def start_pitch(self) -> typing.Optional[NamedPitch]:
        """
        Start pitch of pitch range.

        ..  container:: example

            >>> pitch_range = abjad.PitchRange("[C3, C7]")
            >>> pitch_range.start_pitch
            NamedPitch('c')

        """
        if self._start is None:
            return None
        return self._start[0]

    @property
    def start_pitch_is_included_in_range(self) -> bool:
        """
        Is true when start pitch is included in range.

        ..  container:: example

            >>> pitch_range = abjad.PitchRange("[C3, C7]")
            >>> pitch_range.start_pitch_is_included_in_range
            True

        """
        if self._start is None:
            return True
        return self._start[1] == "inclusive"

    @property
    def stop_pitch(self) -> typing.Optional[NamedPitch]:
        """
        Stop pitch of pitch range.

        ..  container:: example

            >>> pitch_range = abjad.PitchRange("[C3, C7]")
            >>> pitch_range.stop_pitch
            NamedPitch("c''''")

        """
        if self._stop is None:
            return None
        return self._stop[0]

    @property
    def stop_pitch_is_included_in_range(self) -> bool:
        """
        Is true when stop pitch is included in range.

        ..  container:: example

            >>> pitch_range = abjad.PitchRange("[C3, C7]")
            >>> pitch_range.stop_pitch_is_included_in_range
            True

        """
        if self._stop is None:
            return True
        return self._stop[1] == "inclusive"

    ### PUBLIC METHODS ###

    @staticmethod
    def from_octave(octave) -> "PitchRange":
        """
        Initializes pitch range from octave.

        ..  container:: example

            >>> abjad.PitchRange.from_octave(5)
            PitchRange('[C5, C6)')

        """
        octave = Octave(octave)
        return PitchRange(f"[C{octave.number}, C{octave.number + 1})")

    @staticmethod
    def from_pitches(
        start_pitch,
        stop_pitch,
        start_pitch_is_included_in_range=True,
        stop_pitch_is_included_in_range=True,
    ) -> "PitchRange":
        """
        Initializes pitch range from numbers.

        ..  container:: example

            >>> abjad.PitchRange.from_pitches(-18, 19)
            PitchRange('[F#2, G5]')

        """
        if start_pitch is None:
            start_pitch_string = "-inf"
        else:
            start_pitch_string = str(NamedPitch(start_pitch))
        if stop_pitch is None:
            stop_pitch_string = "+inf"
        else:
            stop_pitch_string = str(NamedPitch(stop_pitch))
        start_containment = "["
        if not start_pitch_is_included_in_range:
            start_containment = "("
        stop_containment = "]"
        if not stop_pitch_is_included_in_range:
            stop_containment = ")"
        string = "{}{}, {}{}"
        string = string.format(
            start_containment,
            start_pitch_string,
            stop_pitch_string,
            stop_containment,
        )
        pitch_range = PitchRange(string)
        return pitch_range

    @classmethod
    def is_range_string(class_, argument) -> bool:
        """Is true when ``argument`` is a pitch range string.

        ..  container:: example

            >>> abjad.PitchRange.is_range_string("[A0, C8]")
            True

            >>> abjad.PitchRange.is_range_string("[A#0, Cb~8]")
            True

            >>> abjad.PitchRange.is_range_string("[A#+0, cs'')")
            True

            >>> abjad.PitchRange.is_range_string("(b,,,, ctqs]")
            True

        ..  container:: example

            >>> abjad.PitchRange.is_range_string("text")
            False

        The regex that underlies this predicate matches against two
        comma-separated pitches enclosed in some combination of square
        brackets and round parentheses.
        """
        if not isinstance(argument, str):
            return False
        return bool(_lib._range_string_regex.match(argument))

    def voice_pitch_class(self, pitch_class):
        """
        Voices ``pitch_class``:

        ..  container:: example

            Voices C three times:

            >>> pitch_range = abjad.PitchRange("[C4, C6]")
            >>> pitch_range.voice_pitch_class("c")
            (NamedPitch("c'"), NamedPitch("c''"), NamedPitch("c'''"))

        ..  container:: example

            Voices B two times:

            >>> pitch_range = abjad.PitchRange("[C4, C6]")
            >>> pitch_range.voice_pitch_class("b")
            (NamedPitch("b'"), NamedPitch("b''"))

        ..  container:: example

            Returns empty because B can not voice:

            >>> pitch_range = abjad.PitchRange("[C4, A4)")
            >>> pitch_range.voice_pitch_class('b')
            ()

        """
        named_pitch_class = NamedPitchClass(pitch_class)
        assert self.start_pitch is not None
        assert self.stop_pitch is not None
        pair = (named_pitch_class.name, self.start_pitch.octave.number)
        named_pitch = NamedPitch(pair)
        result = []
        while named_pitch <= self.stop_pitch:
            if named_pitch in self:
                result.append(named_pitch)
            named_pitch += 12
        return tuple(result)



"""
Utilities for typehinting.
"""
import typing

from .duration import Duration
from .ratio import Ratio

IntegerPair = typing.Tuple[int, int]

IntegerSequence = typing.Sequence[int]

DurationTyping = typing.Union[Duration, IntegerPair]

DurationSequenceTyping = typing.Sequence[DurationTyping]

Number = typing.Union[int, float]

NumberPair = typing.Tuple[Number, Number]

PatternTyping = typing.Union[
    typing.Tuple[IntegerSequence], typing.Tuple[IntegerSequence, int]
]

Prototype = typing.Union[typing.Type, typing.Tuple[typing.Type, ...]]

RatioTyping = typing.Union[Duration, Ratio, typing.Tuple[int, ...]]

RatioSequenceTyping = typing.Sequence[RatioTyping]

Strings = typing.Union[str, typing.Sequence[str]]



import copy as python_copy
import itertools

from . import enums, exceptions, get
from .attach import attach, detach
from .duration import Duration
from .indicators.RepeatTie import RepeatTie
from .indicators.Tie import Tie
from .iterate import Iteration
from .makers import NoteMaker
from .pitch.intervals import NamedInterval
from .ratio import Ratio
from .score import BeforeGraceContainer, Chord, Component, Container, Leaf, Note, Tuplet
from .select import Selection
from .sequence import Sequence
from .spanners import tie

### PRIVATE FUNCTIONS ###


def _extract(COMPONENT):
    selection = Selection([COMPONENT])
    parent, start, stop = selection._get_parent_and_start_stop_indices()
    if parent is not None:
        components = list(getattr(COMPONENT, "components", ()))
        parent.__setitem__(slice(start, stop + 1), components)
    return COMPONENT


def _fuse(SELECTION):
    assert SELECTION.are_contiguous_logical_voice()
    if SELECTION.are_leaves():
        return _fuse_leaves(SELECTION)
    elif all(isinstance(_, Tuplet) for _ in SELECTION):
        return _fuse_tuplets(SELECTION)
    else:
        raise Exception(f"can only fuse leaves and tuplets (not {SELECTION}).")


def _fuse_leaves(SELECTION):
    assert SELECTION.are_leaves()
    assert SELECTION.are_contiguous_logical_voice()
    leaves = SELECTION
    if len(leaves) <= 1:
        return leaves
    originally_tied = SELECTION[-1]._has_indicator(Tie)
    total_preprolated = leaves._get_preprolated_duration()
    for leaf in leaves[1:]:
        parent = leaf._parent
        if parent:
            index = parent.index(leaf)
            del parent[index]
    result = _set_leaf_duration(leaves[0], total_preprolated)
    if not originally_tied:
        last_leaf = Selection(result).leaf(-1)
        detach(Tie, last_leaf)
    return result


def _fuse_leaves_by_immediate_parent(SELECTION):
    result = []
    parts = _get_leaves_grouped_by_immediate_parents(SELECTION)
    for part in parts:
        fused = _fuse(part)
        result.append(fused)
    return result


def _fuse_tuplets(SELECTION):
    assert SELECTION.are_contiguous_same_parent(prototype=Tuplet)
    if len(SELECTION) == 0:
        return None
    first = SELECTION[0]
    first_multiplier = first.multiplier
    for tuplet in SELECTION[1:]:
        if tuplet.multiplier != first_multiplier:
            raise ValueError("tuplets must carry same multiplier.")
    assert isinstance(first, Tuplet)
    new_tuplet = Tuplet(first_multiplier, [])
    wrapped = False
    if get.parentage(SELECTION[0]).root is not get.parentage(SELECTION[-1]).root:
        dummy_container = Container(SELECTION)
        wrapped = True
    swap(SELECTION, new_tuplet)
    if wrapped:
        del dummy_container[:]
    return new_tuplet


def _get_leaves_grouped_by_immediate_parents(SELECTION):
    result = []
    pairs_generator = itertools.groupby(SELECTION, lambda x: id(x._parent))
    for key, values_generator in pairs_generator:
        group = Selection(list(values_generator))
        result.append(group)
    return result


def _move_indicators(donor_component, recipient_component):
    for wrapper in get.wrappers(donor_component):
        detach(wrapper, donor_component)
        attach(wrapper, recipient_component)


def _set_leaf_duration(leaf, new_duration):
    new_duration = Duration(new_duration)
    if leaf.multiplier is not None:
        multiplier = new_duration.__div__(leaf.written_duration)
        leaf.multiplier = multiplier
        return Selection(leaf)
    try:
        leaf.written_duration = new_duration
        return Selection(leaf)
    except exceptions.AssignabilityError:
        pass
    maker = NoteMaker()
    components = maker(0, new_duration)
    new_leaves = Selection(components).leaves()
    following_leaf_count = len(new_leaves) - 1
    following_leaves = []
    for i in range(following_leaf_count):
        following_leaf = copy(leaf)
        for indicator in get.indicators(following_leaf):
            if i != following_leaf_count - 1:
                if getattr(indicator, "_time_orientation", enums.Left) != enums.Middle:
                    detach(indicator, following_leaf)
            elif (
                getattr(indicator, "_time_orientation", enums.Left) != enums.Right
                and getattr(indicator, "_time_orientation", enums.Left) != enums.Middle
            ):
                detach(indicator, following_leaf)
        detach(BeforeGraceContainer, following_leaf)
        following_leaves.append(following_leaf)
    if following_leaf_count > 0:
        for indicator in get.indicators(leaf):
            if getattr(indicator, "_time_orientation", enums.Left) == enums.Right:
                detach(indicator, leaf)
    all_leaves = [leaf] + following_leaves
    assert len(all_leaves) == len(new_leaves)
    for all_leaf, new_leaf in zip(all_leaves, new_leaves):
        all_leaf.written_duration = new_leaf.written_duration
    logical_tie = get.logical_tie(leaf)
    logical_tie_leaves = list(logical_tie.leaves)
    for leaf_ in logical_tie:
        detach(Tie, leaf_)
        detach(RepeatTie, leaf_)
    if leaf._parent is not None:
        index = leaf._parent.index(leaf)
        next_ = index + 1
        leaf._parent[next_:next_] = following_leaves
    index = logical_tie_leaves.index(leaf)
    next_ = index + 1
    logical_tie_leaves[next_:next_] = following_leaves
    if 1 < len(logical_tie_leaves) and isinstance(leaf, (Note, Chord)):
        tie(logical_tie_leaves)
    if isinstance(components[0], Leaf):
        return Selection(all_leaves)
    else:
        assert isinstance(components[0], Tuplet)
        assert len(components) == 1
        tuplet = components[0]
        multiplier = tuplet.multiplier
        tuplet = Tuplet(multiplier, [])
        if not isinstance(all_leaves, Selection):
            all_leaves = Selection(all_leaves)
        wrap(all_leaves, tuplet)
        return Selection(tuplet)


def _split_container_at_index(CONTAINER, i):
    """
    Splits container to the left of index ``i``.

    Preserves tuplet multiplier when container is a tuplet.

    Preserves time signature denominator when container is a measure.

    Resizes resizable containers.


    Returns split parts.
    """
    # partition my components
    left_components = CONTAINER[:i]
    right_components = CONTAINER[i:]
    # instantiate new left and right containers
    if isinstance(CONTAINER, Tuplet):
        multiplier = CONTAINER.multiplier
        left = type(CONTAINER)(multiplier, [])
        wrap(left_components, left)
        right = type(CONTAINER)(multiplier, [])
        wrap(right_components, right)
    else:
        left = CONTAINER.__copy__()
        wrap(left_components, left)
        right = CONTAINER.__copy__()
        wrap(right_components, right)
    # save left and right containers together for iteration
    halves = (left, right)
    nonempty_halves = [half for half in halves if len(half)]
    # incorporate left and right parents in score if possible
    selection = Selection(CONTAINER)
    parent, start, stop = selection._get_parent_and_start_stop_indices()
    if parent is not None:
        parent._components.__setitem__(slice(start, stop + 1), nonempty_halves)
        for part in nonempty_halves:
            part._set_parent(parent)
    else:
        left._set_parent(None)
        right._set_parent(None)
    # return new left and right containers
    return halves


def _split_container_by_duration(CONTAINER, duration):
    if CONTAINER.simultaneous:
        return _split_simultaneous_by_duration(CONTAINER, duration=duration)
    duration = Duration(duration)
    assert 0 <= duration, repr(duration)
    if duration == 0:
        # TODO: disallow and raise Exception
        return [], CONTAINER
    # get split point score offset
    timespan = get.timespan(CONTAINER)
    global_split_point = timespan.start_offset + duration
    # get any duration-crossing descendents
    cross_offset = timespan.start_offset + duration
    duration_crossing_descendants = []
    for descendant in get.descendants(CONTAINER):
        timespan = get.timespan(descendant)
        start_offset = timespan.start_offset
        stop_offset = timespan.stop_offset
        if start_offset < cross_offset < stop_offset:
            duration_crossing_descendants.append(descendant)
    # any duration-crossing leaf will be at end of list
    bottom = duration_crossing_descendants[-1]
    did_split_leaf = False
    # if split point necessitates leaf split
    if isinstance(bottom, Leaf):
        assert isinstance(bottom, Leaf)
        original_bottom_parent = bottom._parent
        did_split_leaf = True
        timespan = get.timespan(bottom)
        split_point_in_bottom = global_split_point - timespan.start_offset
        new_leaves = _split_leaf_by_durations(
            bottom,
            [split_point_in_bottom],
        )
        if new_leaves[0]._parent is not original_bottom_parent:
            new_leaves_tuplet_wrapper = new_leaves[0]._parent
            assert isinstance(new_leaves_tuplet_wrapper, Tuplet)
            assert new_leaves_tuplet_wrapper._parent is original_bottom_parent
            _split_container_by_duration(
                new_leaves_tuplet_wrapper,
                split_point_in_bottom,
            )
        for leaf in new_leaves:
            timespan = get.timespan(leaf)
            if timespan.stop_offset == global_split_point:
                leaf_left_of_split = leaf
            if timespan.start_offset == global_split_point:
                leaf_right_of_split = leaf
        duration_crossing_containers = duration_crossing_descendants[:-1]
        assert len(duration_crossing_containers)
    # if split point falls between leaves
    # then find leaf to immediate right of split point
    # in order to start upward crawl through duration-crossing containers
    else:
        duration_crossing_containers = duration_crossing_descendants[:]
        for leaf in Iteration(bottom).leaves():
            timespan = get.timespan(leaf)
            if timespan.start_offset == global_split_point:
                leaf_right_of_split = leaf
                leaf_left_of_split = get.leaf(leaf, -1)
                break
        else:
            raise Exception("can not split empty container {bottom!r}.")
    assert leaf_left_of_split is not None
    assert leaf_right_of_split is not None
    # find component to right of split
    # that is also immediate child of last duration-crossing container
    for component in leaf_right_of_split._get_parentage():
        if component._parent is duration_crossing_containers[-1]:
            highest_level_component_right_of_split = component
            break
    else:
        raise ValueError("should not be able to get here.")
    # crawl back up through duration-crossing containers and split each
    previous = highest_level_component_right_of_split
    for container in reversed(duration_crossing_containers):
        assert isinstance(container, Container)
        index = container.index(previous)
        left, right = _split_container_at_index(container, index)
        previous = right
    # reapply tie here if crawl above killed tie applied to leaves
    if did_split_leaf:
        if isinstance(leaf_left_of_split, Note):
            if (
                get.parentage(leaf_left_of_split).root
                is get.parentage(leaf_right_of_split).root
            ):
                leaves_around_split = (
                    leaf_left_of_split,
                    leaf_right_of_split,
                )
                selection = Selection(leaves_around_split)
                selection._attach_tie_to_leaves()
    # return list-wrapped halves of container
    return [left], [right]


def _split_simultaneous_by_duration(CONTAINER, duration):
    assert CONTAINER.simultaneous
    left_components, right_components = [], []
    for component in CONTAINER[:]:
        halves = _split_container_by_duration(component, duration=duration)
        left_components_, right_components_ = halves
        left_components.extend(left_components_)
        right_components.extend(right_components_)
    left_components = Selection(left_components)
    right_components = Selection(right_components)
    left_container = CONTAINER.__copy__()
    right_container = CONTAINER.__copy__()
    left_container.extend(left_components)
    right_container.extend(right_components)
    if get.parentage(CONTAINER).parent is not None:
        containers = Selection([left_container, right_container])
        replace(CONTAINER, containers)
    # return list-wrapped halves of container
    return [left_container], [right_container]


def _split_leaf_by_durations(LEAF, durations, cyclic=False):
    durations = [Duration(_) for _ in durations]
    durations = Sequence(durations)
    leaf_duration = get.duration(LEAF)
    if cyclic:
        durations = durations.repeat_to_weight(leaf_duration)
    if sum(durations) < leaf_duration:
        last_duration = leaf_duration - sum(durations)
        durations = list(durations)
        durations.append(last_duration)
        durations = Sequence(durations)
    durations = durations.truncate(weight=leaf_duration)
    originally_tied = LEAF._has_indicator(Tie)
    originally_repeat_tied = LEAF._has_indicator(RepeatTie)
    result_selections = []
    # detach grace containers
    before_grace_container = LEAF._before_grace_container
    if before_grace_container is not None:
        detach(before_grace_container, LEAF)
    after_grace_container = LEAF._after_grace_container
    if after_grace_container is not None:
        detach(after_grace_container, LEAF)
    # do other things
    leaf_prolation = get.parentage(LEAF).prolation
    for duration in durations:
        new_leaf = python_copy.copy(LEAF)
        preprolated_duration = duration / leaf_prolation
        selection = _set_leaf_duration(new_leaf, preprolated_duration)
        result_selections.append(selection)
    result_components = Sequence(result_selections).flatten(depth=-1)
    result_components = Selection(result_components)
    result_leaves = Selection(result_components).leaves(grace=False)
    assert all(isinstance(_, Selection) for _ in result_selections)
    assert all(isinstance(_, Component) for _ in result_components)
    assert result_leaves.are_leaves()
    # strip result leaves of all indicators
    for leaf in result_leaves:
        detach(object, leaf)
    # replace leaf with flattened result
    if get.parentage(LEAF).parent is not None:
        replace(LEAF, result_components)
    # move indicators
    first_result_leaf = result_leaves[0]
    last_result_leaf = result_leaves[-1]
    for indicator in get.indicators(LEAF):
        detach(indicator, LEAF)
        direction = getattr(indicator, "_time_orientation", enums.Left)
        if direction is enums.Left:
            attach(indicator, first_result_leaf)
        elif direction == enums.Right:
            attach(indicator, last_result_leaf)
        elif direction == enums.Middle:
            attach(indicator, first_result_leaf)
            indicator_copy = python_copy.copy(indicator)
            attach(indicator_copy, last_result_leaf)
        else:
            raise ValueError(direction)
    # reattach grace containers
    if before_grace_container is not None:
        attach(before_grace_container, first_result_leaf)
    if after_grace_container is not None:
        attach(after_grace_container, last_result_leaf)
    # fuse tuplets
    if isinstance(result_components[0], Tuplet):
        fuse(result_components)
    # tie split notes
    if isinstance(LEAF, (Note, Chord)) and 1 < len(result_leaves):
        result_leaves._attach_tie_to_leaves()
    if originally_repeat_tied and not result_leaves[0]._has_indicator(RepeatTie):
        attach(RepeatTie(), result_leaves[0])
    if originally_tied and not result_leaves[-1]._has_indicator(Tie):
        attach(Tie(), result_leaves[-1])
    assert isinstance(result_leaves, Selection)
    assert all(isinstance(_, Leaf) for _ in result_leaves)
    return result_leaves


### PUBLIC FUNCTIONS ###


def copy(argument, n=1):
    r"""
    Copies client.

    ..  container:: example

        Copies explicit clefs:

        >>> staff = abjad.Staff("c'8 cs'8 d'8 ef'8 e'8 f'8 fs'8 g'8")
        >>> clef = abjad.Clef('treble')
        >>> abjad.attach(clef, staff[0])
        >>> clef = abjad.Clef('bass')
        >>> abjad.attach(clef, staff[4])
        >>> copied_notes = abjad.mutate.copy(staff[:2])
        >>> staff.extend(copied_notes)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "treble"
                c'8
                cs'8
                d'8
                ef'8
                \clef "bass"
                e'8
                f'8
                fs'8
                g'8
                \clef "treble"
                c'8
                cs'8
            }

    ..  container:: example

        Does not copy implicit clefs:

        >>> staff = abjad.Staff("c'8 cs'8 d'8 ef'8 e'8 f'8 fs'8 g'8")
        >>> clef = abjad.Clef('treble')
        >>> abjad.attach(clef, staff[0])
        >>> clef = abjad.Clef('bass')
        >>> abjad.attach(clef, staff[4])
        >>> copied_notes = abjad.mutate.copy(staff[2:4])
        >>> staff.extend(copied_notes)

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "treble"
                c'8
                cs'8
                d'8
                ef'8
                \clef "bass"
                e'8
                f'8
                fs'8
                g'8
                d'8
                ef'8
            }

    ..  container:: example

        Copy components one time:

        >>> staff = abjad.Staff(r"c'8 d'8 e'8 f'8")
        >>> staff.extend(r"g'8 a'8 b'8 c''8")
        >>> time_signature = abjad.TimeSignature((2, 4))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 2/4
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
            }

        >>> selection = staff[2:4]
        >>> result = selection._copy()
        >>> new_staff = abjad.Staff(result)
        >>> abjad.show(new_staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(new_staff)
            >>> print(string)
            \new Staff
            {
                e'8
                f'8
            }

        >>> staff[2] is new_staff[0]
        False

    Returns selection of new components.
    """
    if isinstance(argument, Component):
        selection = Selection(argument)
    else:
        selection = argument
    if n == 1:
        result = selection._copy()
        if isinstance(argument, Component):
            if len(result) == 1:
                result = result[0]
        return result
    else:
        result = []
        for _ in range(n):
            result_ = copy(argument)
            result.append(result_)
        return result


def eject_contents(argument):
    r"""
    Ejects contents from outside-of-score container.

    ..  container:: example

        Ejects leaves from container:

        >>> container = abjad.Container("c'4 ~ c'4 d'4 ~ d'4")
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c'4
                ~
                c'4
                d'4
                ~
                d'4
            }

        Returns container contents as a selection with spanners preserved:

        >>> contents = abjad.mutate.eject_contents(container)
        >>> contents
        Selection([Note("c'4"), Note("c'4"), Note("d'4"), Note("d'4")])

        Container contents can be safely added to a new container:

        >>> staff = abjad.Staff(contents, lilypond_type='RhythmicStaff')
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new RhythmicStaff
            {
                c'4
                ~
                c'4
                d'4
                ~
                d'4
            }

        New container is well formed:

        >>> abjad.wf.wellformed(staff)
        True

        Old container is empty:

        >>> container
        Container()

    Returns container contents as selection.
    """
    return argument._eject_contents()


def extract(argument):
    r"""
    Extracts mutation client from score.

    Leaves children of mutation client in score.

    ..  container:: example

        Extract tuplets:

        >>> staff = abjad.Staff()
        >>> staff.append(abjad.Tuplet((3, 2), "c'4 e'4"))
        >>> staff.append(abjad.Tuplet((3, 2), "d'4 f'4"))
        >>> leaves = abjad.select(staff).leaves()
        >>> time_signature = abjad.TimeSignature((3, 4))
        >>> abjad.attach(time_signature, leaves[0])
        >>> abjad.hairpin('p < f', leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    \time 3/4
                    c'4
                    \p
                    \<
                    e'4
                }
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    d'4
                    f'4
                    \f
                }
            }

        >>> empty_tuplet = abjad.mutate.extract(staff[-1])
        >>> empty_tuplet = abjad.mutate.extract(staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 3/4
                c'4
                \p
                \<
                e'4
                d'4
                f'4
                \f
            }

    ..  container:: example

        Scales tuplet contents and then extracts tuplet:

        >>> staff = abjad.Staff()
        >>> staff.append(abjad.Tuplet((3, 2), "c'4 e'4"))
        >>> staff.append(abjad.Tuplet((3, 2), "d'4 f'4"))
        >>> leaves = abjad.select(staff).leaves()
        >>> abjad.hairpin('p < f', leaves)
        >>> time_signature = abjad.TimeSignature((3, 4))
        >>> abjad.attach(time_signature, leaves[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    \time 3/4
                    c'4
                    \p
                    \<
                    e'4
                }
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    d'4
                    f'4
                    \f
                }
            }

        >>> abjad.mutate.scale(staff[-1], abjad.Multiplier((3, 2)))
        >>> empty_tuplet = abjad.mutate.extract(staff[-1])
        >>> abjad.mutate.scale(staff[0], abjad.Multiplier((3, 2)))
        >>> empty_tuplet = abjad.mutate.extract(staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 3/4
                c'4.
                \p
                \<
                e'4.
                d'4.
                f'4.
                \f
            }

    ..  container:: example

        Extracting out-of-score component does nothing and returns
        component:

        >>> tuplet = abjad.Tuplet((3, 2), "c'4 e'4")
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \times 3/2 {
                c'4
                e'4
            }

        >>> abjad.mutate.extract(tuplet)
        Tuplet('2:3', "c'4 e'4")

        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \times 3/2 {
                c'4
                e'4
            }

    Returns mutation client.
    """
    return _extract(argument)


def fuse(argument):
    r"""
    Fuses mutation client.

    ..  container:: example

        Fuses in-score leaves:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.mutate.fuse(staff[1:])
        Selection([Note("d'4.")])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                d'4.
            }

    ..  container:: example

        Fuses parent-contiguous tuplets in selection:

        >>> tuplet_1 = abjad.Tuplet((2, 3), "c'8 d' e'")
        >>> abjad.beam(tuplet_1[:])
        >>> tuplet_2 = abjad.Tuplet((2, 3), "c'16 d' e'")
        >>> abjad.slur(tuplet_2[:])
        >>> staff = abjad.Staff([tuplet_1, tuplet_2])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \times 2/3 {
                    c'8
                    [
                    d'8
                    e'8
                    ]
                }
                \times 2/3 {
                    c'16
                    (
                    d'16
                    e'16
                    )
                }
            }

        >>> tuplets = staff[:]
        >>> abjad.mutate.fuse(tuplets)
        Tuplet('3:2', "c'8 d'8 e'8 c'16 d'16 e'16")
        >>> abjad.show(staff) #doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \times 2/3 {
                    c'8
                    [
                    d'8
                    e'8
                    ]
                    c'16
                    (
                    d'16
                    e'16
                    )
                }
            }

        Returns new tuplet in selection.

        Fuses zero or more parent-contiguous ``tuplets``.

        Allows in-score ``tuplets``.

        Allows outside-of-score ``tuplets``.

        All ``tuplets`` must carry the same multiplier.

        All ``tuplets`` must be of the same type.

    ..  container:: example

        REGRESSION. Trims tie from fused note:

        >>> staff = abjad.Staff("d'8 ~ d'32 ~ d'16 d'32")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                d'8
                ~
                d'32
                ~
                d'16
                d'32
            }

        >>> logical_tie = abjad.select(staff[0]).logical_tie()
        >>> abjad.mutate.fuse(logical_tie)
        Selection([Note("d'8..")])

        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                d'8..
                d'32
            }

        >>> abjad.get.has_indicator(staff[0], abjad.Tie)
        False

    Returns selection.
    """
    if isinstance(argument, Component):
        selection = Selection(argument)
        return _fuse(selection)
    elif isinstance(argument, Selection) and argument.are_contiguous_logical_voice():
        selection = Selection(argument)
        return _fuse(selection)


def logical_tie_to_tuplet(argument, proportions) -> Tuplet:
    r"""
    Changes logical tie to tuplet.

    ..  container:: example

        >>> staff = abjad.Staff(r"df'8 c'8 ~ c'16 cqs''4")
        >>> abjad.attach(abjad.Dynamic('p'), staff[0])
        >>> abjad.attach(abjad.StartHairpin('<'), staff[0])
        >>> abjad.attach(abjad.Dynamic('f'), staff[-1])
        >>> abjad.override(staff).DynamicLineSpanner.staff_padding = 3
        >>> time_signature = abjad.TimeSignature((9, 16))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 3
            }
            {
                \time 9/16
                df'8
                \p
                \<
                c'8
                ~
                c'16
                cqs''4
                \f
            }

        >>> logical_tie = abjad.select(staff[1]).logical_tie()
        >>> abjad.mutate.logical_tie_to_tuplet(logical_tie, [2, 1, 1, 1])
        Tuplet('5:3', "c'8 c'16 c'16 c'16")

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 3
            }
            {
                \time 9/16
                df'8
                \p
                \<
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/5 {
                    c'8
                    c'16
                    c'16
                    c'16
                }
                cqs''4
                \f
            }

        >>> abjad.show(staff) # doctest: +SKIP

    ..  container:: example

        >>> staff = abjad.Staff(r"c'8 ~ c'16 cqs''4")
        >>> abjad.hairpin('p < f', staff[:])
        >>> abjad.override(staff).DynamicLineSpanner.staff_padding = 3
        >>> time_signature = abjad.TimeSignature((7, 16))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 3
            }
            {
                \time 7/16
                c'8
                \p
                \<
                ~
                c'16
                cqs''4
                \f
            }

    """
    proportions = Ratio(proportions)
    target_duration = argument._get_preprolated_duration()
    prolated_duration = target_duration / sum(proportions.numbers)
    basic_written_duration = prolated_duration.equal_or_greater_power_of_two
    written_durations = [_ * basic_written_duration for _ in proportions.numbers]
    maker = NoteMaker()
    try:
        notes = Selection([Note(0, _) for _ in written_durations])
    except exceptions.AssignabilityError:
        denominator = target_duration._denominator
        note_durations = [Duration(_, denominator) for _ in proportions.numbers]
        notes = maker(0, note_durations)
    tuplet = Tuplet.from_duration(target_duration, notes)
    for leaf in argument:
        detach(Tie, leaf)
        detach(RepeatTie, leaf)
    replace(argument, tuplet)
    return tuplet


def replace(argument, recipients, wrappers=False):
    r"""
    Replaces mutation client (and contents of mutation client) with
    ``recipients``.

    ..  container:: example

        Replaces in-score tuplet (and children of tuplet) with notes.
        Functions exactly the same as container setitem:

        >>> tuplet_1 = abjad.Tuplet((2, 3), "c'4 d'4 e'4")
        >>> tuplet_2 = abjad.Tuplet((2, 3), "d'4 e'4 f'4")
        >>> staff = abjad.Staff([tuplet_1, tuplet_2])
        >>> leaves = abjad.select(staff).leaves()
        >>> abjad.hairpin('p < f', leaves)
        >>> abjad.slur(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \times 2/3 {
                    c'4
                    \p
                    \<
                    (
                    d'4
                    e'4
                }
                \times 2/3 {
                    d'4
                    e'4
                    f'4
                    \f
                    )
                }
            }

        >>> maker = abjad.NoteMaker()
        >>> notes = maker("c' d' e' f' c' d' e' f'", (1, 16))
        >>> abjad.mutate.replace([tuplet_1], notes)
        >>> abjad.attach(abjad.Dynamic('p'), staff[0])
        >>> abjad.attach(abjad.StartHairpin('<'), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'16
                \p
                \<
                d'16
                e'16
                f'16
                c'16
                d'16
                e'16
                f'16
                \times 2/3 {
                    d'4
                    e'4
                    f'4
                    \f
                    )
                }
            }

        Preserves both hairpin and slur.

    ..  container:: example

        Copies no wrappers when ``wrappers`` is false:

        >>> staff = abjad.Staff("c'2 f'4 g'")
        >>> abjad.attach(abjad.Clef('alto'), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "alto"
                c'2
                f'4
                g'4
            }

        >>> for leaf in staff:
        ...     leaf, abjad.get.effective(leaf, abjad.Clef)
        ...
        (Note("c'2"), Clef('alto'))
        (Note("f'4"), Clef('alto'))
        (Note("g'4"), Clef('alto'))

        >>> chord = abjad.Chord("<d' e'>2")
        >>> abjad.mutate.replace(staff[0], chord)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <d' e'>2
                f'4
                g'4
            }

        >>> for leaf in staff:
        ...     leaf, abjad.get.effective(leaf, abjad.Clef)
        ...
        (Chord("<d' e'>2"), None)
        (Note("f'4"), None)
        (Note("g'4"), None)

        >>> abjad.wf.wellformed(staff)
        True

    ..  container:: example

        Set ``wrappers`` to true to copy all wrappers from one leaf to
        another leaf (and avoid full-score update). Only works from one
        leaf to another leaf:

        >>> staff = abjad.Staff("c'2 f'4 g'")
        >>> abjad.attach(abjad.Clef('alto'), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "alto"
                c'2
                f'4
                g'4
            }

        >>> for leaf in staff:
        ...     leaf, abjad.get.effective(leaf, abjad.Clef)
        ...
        (Note("c'2"), Clef('alto'))
        (Note("f'4"), Clef('alto'))
        (Note("g'4"), Clef('alto'))

        >>> chord = abjad.Chord("<d' e'>2")
        >>> abjad.mutate.replace(staff[0], chord, wrappers=True)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "alto"
                <d' e'>2
                f'4
                g'4
            }

        >>> for leaf in staff:
        ...     leaf, abjad.get.effective(leaf, abjad.Clef)
        ...
        (Chord("<d' e'>2"), Clef('alto'))
        (Note("f'4"), Clef('alto'))
        (Note("g'4"), Clef('alto'))

        >>> abjad.wf.wellformed(staff)
        True

    ..  container:: example

        ..  todo:: Fix.

        Introduces duplicate ties:

        >>> staff = abjad.Staff("c'2 ~ c'2")
        >>> maker = abjad.NoteMaker()
        >>> tied_notes = maker(0, abjad.Duration(5, 8))
        >>> abjad.mutate.replace(staff[:1], tied_notes)

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            c'2
            ~
            c'8
            c'2
        }

    Returns none.
    """
    if isinstance(argument, Selection):
        donors = argument
    else:
        donors = Selection(argument)
    assert donors.are_contiguous_same_parent()
    if not isinstance(recipients, Selection):
        recipients = Selection(recipients)
    assert recipients.are_contiguous_same_parent()
    if not donors:
        return
    if wrappers is True:
        if 1 < len(donors) or not isinstance(donors[0], Leaf):
            raise Exception(f"set wrappers only with single leaf: {donors!r}.")
        if 1 < len(recipients) or not isinstance(recipients[0], Leaf):
            raise Exception(f"set wrappers only with single leaf: {recipients!r}.")
        donor = donors[0]
        wrappers = get.wrappers(donor)
        recipient = recipients[0]
    parent, start, stop = donors._get_parent_and_start_stop_indices()
    assert parent is not None, repr(donors)
    parent.__setitem__(slice(start, stop + 1), recipients)
    if not wrappers:
        return
    for wrapper in wrappers:
        # bypass Wrapper._bind_component()
        # to avoid full-score update / traversal;
        # this works because one-to-one leaf replacement
        # including all (persistent) indicators
        # doesn't change score structure:
        donor._wrappers.remove(wrapper)
        wrapper._component = recipient
        recipient._wrappers.append(wrapper)
        context = wrapper._find_correct_effective_context()
        if context is not None:
            context._dependent_wrappers.append(wrapper)


def scale(argument, multiplier) -> None:
    r"""
    Scales mutation client by ``multiplier``.

    ..  container:: example

        Scales note duration by dot-generating multiplier:

        >>> staff = abjad.Staff("c'8 ( d'8 e'8 f'8 )")
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.mutate.scale(staff[1], abjad.Multiplier(3, 2))
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                (
                d'8.
                e'8
                f'8
                )
            }

    ..  container:: example

        Scales tied leaves by dot-generating mutliplier:

        >>> staff = abjad.Staff(r"c'8 \accent ~ c'8 d'8")
        >>> time_signature = abjad.TimeSignature((3, 8))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 3/8
                c'8
                - \accent
                ~
                c'8
                d'8
            }

        >>> logical_tie = abjad.select(staff[0]).logical_tie()
        >>> logical_tie = abjad.mutate.scale(logical_tie, abjad.Multiplier(3, 2))
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 3/8
                c'8.
                - \accent
                ~
                c'8.
                d'8
            }

    ..  container:: example

        Scales leaves in container by dot-generating multiplier:

        >>> container = abjad.Container(r"c'8 ( d'8 e'8 f'8 )")
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c'8
                (
                d'8
                e'8
                f'8
                )
            }

        >>> abjad.mutate.scale(container, abjad.Multiplier(3, 2))
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c'8.
                (
                d'8.
                e'8.
                f'8.
                )
            }

    ..  container:: example

        Scales leaves in tuplet:

        >>> staff = abjad.Staff()
        >>> tuplet = abjad.Tuplet((4, 5), "c'8 d'8 e'8 f'8 g'8")
        >>> staff.append(tuplet)
        >>> time_signature = abjad.TimeSignature((4, 8))
        >>> leaf = abjad.get.leaf(staff, 0)
        >>> abjad.attach(time_signature, leaf)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \times 4/5 {
                    \time 4/8
                    c'8
                    d'8
                    e'8
                    f'8
                    g'8
                }
            }

        >>> abjad.mutate.scale(tuplet, abjad.Multiplier(2))
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \times 4/5 {
                    \time 4/8
                    c'4
                    d'4
                    e'4
                    f'4
                    g'4
                }
            }

    ..  container:: example

        Scales leaves carrying LilyPond multiplier:

        >>> note = abjad.Note("c'8", multiplier=(1, 2))
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(note)
            >>> print(string)
            c'8 * 1/2

        >>> abjad.mutate.scale(note, abjad.Multiplier(1, 2))
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(note)
            >>> print(string)
            c'16 * 1/2

    """
    if hasattr(argument, "_scale"):
        argument._scale(multiplier)
    else:
        assert isinstance(argument, Selection)
        for component in argument:
            component._scale(multiplier)


# TODO: add tests of tupletted notes and rests.
# TODO: add examples that show indicator handling.
# TODO: add example showing grace and after grace handling.
def split(argument, durations, cyclic=False):
    r"""
    Splits mutation client by ``durations``.

    ..  container:: example

        Splits leaves cyclically and ties split notes:

        >>> staff = abjad.Staff("c'1 d'1")
        >>> abjad.hairpin('p < f', staff[:])
        >>> abjad.override(staff).DynamicLineSpanner.staff_padding = 3
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 3
            }
            {
                c'1
                \p
                \<
                d'1
                \f
            }

        >>> durations = [(3, 4)]
        >>> result = abjad.mutate.split(
        ...     staff[:],
        ...     durations,
        ...     cyclic=True,
        ...     )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 3
            }
            {
                c'2.
                \p
                ~
                c'4
                \<
                d'2
                \f
                ~
                d'2
            }

    ..  container:: example

        Splits custom voice and preserves context name:

        >>> voice = abjad.Voice(
        ...     "c'4 d' e' f'",
        ...     lilypond_type='CustomVoice',
        ...     name='1',
        ...     )
        >>> staff = abjad.Staff([voice])
        >>> abjad.hairpin('p < f', voice[:])
        >>> abjad.override(staff).DynamicLineSpanner.staff_padding = 3
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 3
            }
            {
                \context CustomVoice = "1"
                {
                    c'4
                    \p
                    \<
                    d'4
                    e'4
                    f'4
                    \f
                }
            }

        >>> durations = [(1, 8)]
        >>> result = abjad.mutate.split(
        ...     staff[:],
        ...     durations,
        ...     cyclic=True,
        ...     )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 3
            }
            {
                \context CustomVoice = "1"
                {
                    c'8
                    \p
                    ~
                }
                \context CustomVoice = "1"
                {
                    c'8
                    \<
                }
                \context CustomVoice = "1"
                {
                    d'8
                    ~
                }
                \context CustomVoice = "1"
                {
                    d'8
                }
                \context CustomVoice = "1"
                {
                    e'8
                    ~
                }
                \context CustomVoice = "1"
                {
                    e'8
                }
                \context CustomVoice = "1"
                {
                    f'8
                    \f
                    ~
                }
                \context CustomVoice = "1"
                {
                    f'8
                }
            }

        >>> for voice in staff:
        ...     voice
        ...
        Voice("c'8", lilypond_type='CustomVoice', name='1')
        Voice("c'8", lilypond_type='CustomVoice', name='1')
        Voice("d'8", lilypond_type='CustomVoice', name='1')
        Voice("d'8", lilypond_type='CustomVoice', name='1')
        Voice("e'8", lilypond_type='CustomVoice', name='1')
        Voice("e'8", lilypond_type='CustomVoice', name='1')
        Voice("f'8", lilypond_type='CustomVoice', name='1')
        Voice("f'8", lilypond_type='CustomVoice', name='1')

    ..  container:: example

        Splits parallel container:

        >>> voice_1 = abjad.Voice(
        ...     "e''4 ( es'' f'' fs'' )",
        ...     name='Voice_1',
        ...     )
        >>> voice_2 = abjad.Voice(
        ...     r"c'4 \p \< cs' d' ds' \f",
        ...     name='Voice_2',
        ...     )
        >>> abjad.override(voice_1).Stem.direction = abjad.Up
        >>> abjad.override(voice_1).Slur.direction = abjad.Up
        >>> container = abjad.Container(
        ...     [voice_1, voice_2],
        ...     simultaneous=True,
        ...     )
        >>> abjad.override(voice_2).Stem.direction = abjad.Down
        >>> staff = abjad.Staff([container])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <<
                    \context Voice = "Voice_1"
                    \with
                    {
                        \override Slur.direction = #up
                        \override Stem.direction = #up
                    }
                    {
                        e''4
                        (
                        es''4
                        f''4
                        fs''4
                        )
                    }
                    \context Voice = "Voice_2"
                    \with
                    {
                        \override Stem.direction = #down
                    }
                    {
                        c'4
                        \p
                        \<
                        cs'4
                        d'4
                        ds'4
                        \f
                    }
                >>
            }

        >>> durations = [(3, 8)]
        >>> result = abjad.mutate.split(
        ...     container,
        ...     durations,
        ...     cyclic=False,
        ...     )
        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            <<
                \context Voice = "Voice_1"
                \with
                {
                    \override Slur.direction = #up
                    \override Stem.direction = #up
                }
                {
                    e''4
                    (
                    es''8
                    ~
                }
                \context Voice = "Voice_2"
                \with
                {
                    \override Stem.direction = #down
                }
                {
                    c'4
                    \p
                    \<
                    cs'8
                    ~
                }
            >>
            <<
                \context Voice = "Voice_1"
                \with
                {
                    \override Slur.direction = #up
                    \override Stem.direction = #up
                }
                {
                    es''8
                    f''4
                    fs''4
                    )
                }
                \context Voice = "Voice_2"
                \with
                {
                    \override Stem.direction = #down
                }
                {
                    cs'8
                    d'4
                    ds'4
                    \f
                }
            >>
        }

    ..  container:: example

        Splits leaves with articulations:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Articulation('^'), staff[0])
        >>> abjad.attach(abjad.LaissezVibrer(), staff[1])
        >>> abjad.attach(abjad.Articulation('^'), staff[2])
        >>> abjad.attach(abjad.LaissezVibrer(), staff[3])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \marcato
                d'4
                \laissezVibrer
                e'4
                - \marcato
                f'4
                \laissezVibrer
            }

        >>> durations = [(1, 8)]
        >>> result = abjad.mutate.split(
        ...     staff[:],
        ...     durations,
        ...     cyclic=True,
        ...     )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                - \marcato
                ~
                c'8
                d'8
                ~
                d'8
                \laissezVibrer
                e'8
                - \marcato
                ~
                e'8
                f'8
                ~
                f'8
                \laissezVibrer
            }

    ..  container:: example

        REGRESSION. Preserves tie:

        >>> staff = abjad.Staff("d'2 ~ d'")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                d'2
                ~
                d'2
            }

        >>> result = abjad.mutate.split(staff[0], [(1, 32)])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                d'32
                ~
                d'4...
                ~
                d'2
            }

    Returns list of selections.
    """
    components = argument
    if isinstance(components, Component):
        components = Selection(components)
    assert all(isinstance(_, Component) for _ in components)
    if not isinstance(components, Selection):
        components = Selection(components)
    durations = [Duration(_) for _ in durations]
    assert len(durations), repr(durations)
    total_component_duration = get.duration(components)
    total_split_duration = sum(durations)
    if cyclic:
        durations = Sequence(durations)
        durations = durations.repeat_to_weight(total_component_duration)
        durations = list(durations)
    elif total_split_duration < total_component_duration:
        final_offset = total_component_duration - sum(durations)
        durations.append(final_offset)
    elif total_component_duration < total_split_duration:
        weight = total_component_duration
        durations = Sequence(durations).truncate(weight=weight)
        durations = list(durations)
    # keep copy of durations to partition result components
    durations_copy = durations[:]
    total_split_duration = sum(durations)
    assert total_split_duration == total_component_duration
    result, shard = [], []
    offset_index = 0
    current_shard_duration = Duration(0)
    remaining_components = list(components[:])
    advance_to_next_offset = True
    # build shards:
    # grab next component and next duration each time through loop
    while True:
        # grab next split point
        if advance_to_next_offset:
            if not durations:
                break
            next_split_point = durations.pop(0)
        advance_to_next_offset = True
        # grab next component from input stack of components
        if not remaining_components:
            break
        current_component = remaining_components.pop(0)
        # find where current component endpoint will position us
        duration_ = get.duration(current_component)
        candidate_shard_duration = current_shard_duration + duration_
        # if current component would fill current shard exactly
        if candidate_shard_duration == next_split_point:
            shard.append(current_component)
            result.append(shard)
            shard = []
            current_shard_duration = Duration(0)
            offset_index += 1
        # if current component would exceed current shard
        elif next_split_point < candidate_shard_duration:
            local_split_duration = next_split_point
            local_split_duration -= current_shard_duration
            if isinstance(current_component, Leaf):
                leaf_split_durations = [local_split_duration]
                duration_ = get.duration(current_component)
                current_duration = duration_
                additional_required_duration = current_duration
                additional_required_duration -= local_split_duration
                split_durations = Sequence(durations)
                split_durations = split_durations.split(
                    [additional_required_duration],
                    cyclic=False,
                    overhang=True,
                )
                split_durations = [list(_) for _ in split_durations]
                additional_durations = split_durations[0]
                leaf_split_durations.extend(additional_durations)
                durations = split_durations[-1]
                leaf_shards = _split_leaf_by_durations(
                    current_component,
                    leaf_split_durations,
                    cyclic=False,
                )
                shard.extend(leaf_shards)
                result.append(shard)
                offset_index += len(additional_durations)
            else:
                assert isinstance(current_component, Container)
                pair = _split_container_by_duration(
                    current_component,
                    local_split_duration,
                )
                left_list, right_list = pair
                shard.extend(left_list)
                result.append(shard)
                remaining_components.__setitem__(slice(0, 0), right_list)
            shard = []
            offset_index += 1
            current_shard_duration = Duration(0)
        # if current component would not fill current shard
        elif candidate_shard_duration < next_split_point:
            shard.append(current_component)
            duration_ = get.duration(current_component)
            current_shard_duration += duration_
            advance_to_next_offset = False
        else:
            message = "can not process candidate duration: {!r}."
            message = message.format(candidate_shard_duration)
            raise ValueError(message)
    # append any stub shard
    if len(shard):
        result.append(shard)
    # append any unexamined components
    if len(remaining_components):
        result.append(remaining_components)
    # partition split components according to input durations
    result = Sequence(result).flatten(depth=-1)
    result = Selection(result).partition_by_durations(durations_copy, fill=enums.Exact)
    # return list of shards
    assert all(isinstance(_, Selection) for _ in result)
    return result


def swap(argument, container):
    r"""
    Swaps mutation client for empty ``container``.

    ..  container:: example

        Swaps containers for tuplet:

        >>> staff = abjad.Staff()
        >>> staff.append(abjad.Container("c'4 d'4 e'4"))
        >>> staff.append(abjad.Container("d'4 e'4 f'4"))
        >>> abjad.attach(abjad.TimeSignature((3, 4)), staff[0][0])
        >>> leaves = abjad.select(staff).leaves()
        >>> abjad.hairpin('p < f', leaves)
        >>> measures = staff[:]
        >>> abjad.slur(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                {
                    \time 3/4
                    c'4
                    \p
                    \<
                    (
                    d'4
                    e'4
                }
                {
                    d'4
                    e'4
                    f'4
                    \f
                    )
                }
            }

        >>> containers = staff[:]
        >>> tuplet = abjad.Tuplet((2, 3), [])
        >>> tuplet.denominator = 4
        >>> abjad.mutate.swap(containers, tuplet)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \times 4/6 {
                    \time 3/4
                    c'4
                    \p
                    \<
                    (
                    d'4
                    e'4
                    d'4
                    e'4
                    f'4
                    \f
                    )
                }
            }

    Returns none.
    """
    if isinstance(argument, Selection):
        donors = argument
    else:
        donors = Selection(argument)
    assert donors.are_contiguous_same_parent()
    assert isinstance(container, Container)
    assert not container, repr(container)
    donors._give_components_to_empty_container(container)
    # donors._give_dominant_spanners([container])
    donors._give_position_in_parent_to_container(container)


def transpose(argument, interval):
    r"""
    Transposes notes and chords in ``argument`` by ``interval``.

    ..  todo:: Move to abjad.pitch package.

    ..  container:: example

        Transposes notes and chords in staff:

        >>> staff = abjad.Staff()
        >>> staff.extend("c'4 d'4 e'4 r4")
        >>> abjad.attach(abjad.TimeSignature((4, 4)), staff[0])
        >>> staff.extend("d'4 e'4 <f' a' c''>4")
        >>> abjad.attach(abjad.TimeSignature((3, 4)), staff[4])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                r4
                \time 3/4
                d'4
                e'4
                <f' a' c''>4
            }

        >>> abjad.mutate.transpose(staff, "+m3")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 4/4
                ef'4
                f'4
                g'4
                r4
                \time 3/4
                f'4
                g'4
                <af' c'' ef''>4
            }

    Returns none.
    """
    named_interval = NamedInterval(interval)
    for x in Iteration(argument).components((Note, Chord)):
        if isinstance(x, Note):
            old_written_pitch = x.note_head.written_pitch
            new_written_pitch = old_written_pitch.transpose(named_interval)
            x.note_head.written_pitch = new_written_pitch
        else:
            for note_head in x.note_heads:
                old_written_pitch = note_head.written_pitch
                new_written_pitch = old_written_pitch.transpose(named_interval)
                note_head.written_pitch = new_written_pitch


def wrap(argument, container):
    r"""
    Wraps mutation client in empty ``container``.

    ..  container:: example

        Wraps in-score notes in tuplet:

        >>> staff = abjad.Staff("c'8 [ ( d' e' ] ) c' [ ( d' e' ] )")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                [
                (
                d'8
                e'8
                )
                ]
                c'8
                [
                (
                d'8
                e'8
                )
                ]
            }

        >>> tuplet = abjad.Tuplet((2, 3), [])
        >>> abjad.mutate.wrap(staff[-3:], tuplet)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                [
                (
                d'8
                e'8
                )
                ]
                \times 2/3 {
                    c'8
                    [
                    (
                    d'8
                    e'8
                    )
                    ]
                }
            }

    ..  container:: example

        Wraps outside-score notes in tuplet:

        >>> maker = abjad.NoteMaker()
        >>> notes = maker([0, 2, 4], [(1, 8)])
        >>> tuplet = abjad.Tuplet((2, 3), [])
        >>> abjad.mutate.wrap(notes, tuplet)
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \times 2/3 {
                c'8
                d'8
                e'8
            }

        (This usage merely substitutes for the tuplet initializer.)

    ..  container:: example

        Wraps leaves in container:

        >>> notes = [abjad.Note(n, (1, 8)) for n in range(8)]
        >>> staff = abjad.Staff(notes)
        >>> abjad.attach(abjad.TimeSignature((4, 8)), staff[0])
        >>> container = abjad.Container()
        >>> abjad.mutate.wrap(staff[:4], container)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                {
                    \time 4/8
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
                e'8
                f'8
                fs'8
                g'8
            }

    ..  container:: example

        Wraps each leaf in tuplet:

        >>> notes = [abjad.Note(n, (1, 1)) for n in range(4)]
        >>> staff = abjad.Staff(notes)
        >>> for note in staff:
        ...     tuplet = abjad.Tuplet((2, 3))
        ...     abjad.mutate.wrap(note, tuplet)
        ...

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \tweak edge-height #'(0.7 . 0)
                \times 2/3 {
                    c'1
                }
                \tweak edge-height #'(0.7 . 0)
                \times 2/3 {
                    cs'1
                }
                \tweak edge-height #'(0.7 . 0)
                \times 2/3 {
                    d'1
                }
                \tweak edge-height #'(0.7 . 0)
                \times 2/3 {
                    ef'1
                }
            }

    ..  container:: example

        Raises exception on nonempty ``container``:

        >>> import pytest
        >>> staff = abjad.Staff("c'8 [ ( d' e' ] ) c' [ ( d' e' ] )")
        >>> tuplet = abjad.Tuplet((2, 3), "g'8 a' fs'")
        >>> abjad.mutate.wrap(staff[-3:], tuplet)
        Traceback (most recent call last):
            ...
        Exception: must be empty container: Tuplet('3:2', "g'8 a'8 fs'8").

    ..  container:: example

        REGRESSION. Contexted indicators (like time signature) survive
        wrap:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> leaves = abjad.select(staff).leaves()
        >>> abjad.attach(abjad.TimeSignature((3, 8)), leaves[0])
        >>> container = abjad.Container()
        >>> abjad.mutate.wrap(leaves, container)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                {
                    \time 3/8
                    c'4
                    d'4
                    e'4
                    f'4
                }
            }

        >>> prototype = abjad.TimeSignature
        >>> for component in abjad.iterate(staff).components():
        ...     time_signature = abjad.get.effective(component, prototype)
        ...     print(component, time_signature)
        ...
        <Staff{1}> 3/8
        Container("c'4 d'4 e'4 f'4") 3/8
        c'4 3/8
        d'4 3/8
        e'4 3/8
        f'4 3/8

    Returns none.
    """
    if not isinstance(container, Container) or 0 < len(container):
        raise Exception(f"must be empty container: {container!r}.")
    if isinstance(argument, Component):
        selection = Selection(argument)
    else:
        selection = argument
    assert isinstance(selection, Selection), repr(selection)
    parent, start, stop = selection._get_parent_and_start_stop_indices(
        ignore_before_after_grace=True
    )
    if not selection.are_contiguous_logical_voice(ignore_before_after_grace=True):
        message = "must be contiguous components in same logical voice:\n"
        message += f"   {selection!r}."
        raise Exception(message)
    container._components = list(selection)
    container[:]._set_parents(container)
    if parent is not None:
        parent._components.insert(start, container)
        container._set_parent(parent)
    for component in selection:
        for wrapper in component._wrappers:
            wrapper._effective_context = None
            wrapper._update_effective_context()



from ..bundle import LilyPondFormatBundle
from ..storage import StorageFormatManager


class LilyPondComment:
    r"""
    LilyPond comment.

    ..  container:: example

        Two-word comment:

        >>> note = abjad.Note("c'4")
        >>> comment = abjad.LilyPondComment('a comment')
        >>> abjad.attach(comment, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(note)
            >>> print(string)
            % a comment
            c'4

    ..  container:: example

        Three-word comment:

        >>> note = abjad.Note("c'4")
        >>> comment = abjad.LilyPondComment('yet another comment')
        >>> abjad.attach(comment, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(note)
            >>> print(string)
            % yet another comment
            c'4

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_string", "_format_slot")

    _allowable_format_slots = ("after", "before", "closing", "opening")

    _can_attach_to_containers = True

    _format_leaf_children = False

    ### INITIALIZER ###

    def __init__(self, string: str = None, format_slot: str = "before") -> None:
        if isinstance(string, type(self)):
            argument = string
            string = argument.string
            format_slot = format_slot or argument.format_slot
        else:
            string = str(string)
        self._string = string
        assert format_slot in self._allowable_format_slots, repr(format_slot)
        self._format_slot = format_slot

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __str__(self) -> str:
        """
        Gets string representation of LilyPond comment.

        ..  container:: example

            Two-word comment:

            >>> comment = abjad.LilyPondComment('a comment')
            >>> str(comment)
            '% a comment'

        ..  container:: example

            Three-word comment:

            >>> comment = abjad.LilyPondComment('yet another comment')
            >>> str(comment)
            '% yet another comment'

        """
        return rf"% {self.string}"

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        format_slot = bundle.get(self.format_slot)
        format_slot.comments.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def format_slot(self) -> str:
        """
        Format slot of LilyPond comment.

        ..  container:: example

            Two-word comment:

            >>> comment = abjad.LilyPondComment('a comment')
            >>> comment.format_slot
            'before'

        ..  container:: example

            Three-word comment:

            >>> comment = abjad.LilyPondComment('yet another comment')
            >>> comment.format_slot
            'before'

        """
        return self._format_slot

    @property
    def string(self) -> str:
        """
        Gets string.

        ..  container:: example

            Two-word comment:

            >>> comment = abjad.LilyPondComment('a comment')
            >>> comment.string
            'a comment'

        ..  container:: example

            Three-word comment:

            >>> comment = abjad.LilyPondComment('yet another comment')
            >>> comment.string
            'yet another comment'

        """
        return self._string

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on LilyPond comment.
        """
        pass
